/****************************************************************************
 * NCSA HDF                                                                 *
 * Software Development Group                                               *
 * National Center for Supercomputing Applications                          *
 * University of Illinois at Urbana-Champaign                               *
 * 605 E. Springfield, Champaign IL 61820                                   *
 *                                                                          *
 * For conditions of distribution and use, see the accompanying             *
 * COPYING file.                                                            *
 *                                                                          *
 ****************************************************************************/
#ifdef RCSID
static char RcsId[] = "@(#)sdb_pal.c,v 1.21 1996/04/15 17:58:08 georgev Exp";
#endif

/* sdb_pal.c,v 1.21 1996/04/15 17:58:08 georgev Exp */

/*
   LIST Functions
   --------------
      add_pal      - add palette to palette list
      pal_ann2html - Iterator function to print annotation of palette
      pr_pallabel - Iterator fcn to print labels of palette
      pr_paldesc - Iterator fcn to print descriptions of raster
      pr_palette - Iterator fcn to print palette
      set_lone_pal - Iterator fcn to set this is a lone palette 
      lone_pal     - Iterator fcn to indicate whether this is a lone palette  
      free_pal     - frees palette element 
      cleanup_pal  - frees palette lists

   Regular Functions
   -----------------
      do_pals   - process palettes in file and add to global list
      dump_pals -  print out the info for Palettes in the file 

 */

/* Include our stuff */
#include "sdb.h"
#include "sdb_pal.h"
#include "sdb_ann.h"
#include "sdb_util.h"
#include "glist.h"
#include "htmlstrings.h"
#include "html_util.h"

/*------------------------------------------------------------------------- 
 NAME
      add_pal - add palette to palette list
 DESCRIPTION
      This adds palette and its info to the palette List.
 RETURNS
      SUCCEED/FAIL
--------------------------------------------------------------------------*/
int
add_pal(Generic_list *pal_list, /* palette list */
        uint16 ref,             /* ref of palette to add */
        char *pal,              /* palette itself */
        lvar_st *l_vars)
{
    tpal_st *pal_elem = NULL;
    int    ret_value = SUCCEED;

    /* allocate space for palette element */
    if ((pal_elem = (tpal_st *)HDmalloc(sizeof(tpal_st))) == NULL)
      {
          gateway_err(l_vars->hfp,"add_pal: failed to allocate space for pal_elem \n",0);
          ret_value = FAIL; 
          goto done;
      }   

    /* fill in relevant info */
    pal_elem->ref = ref;
    pal_elem->lone_pal = 1; /* yes for now */

    /* copy palette */
    HDmemcpy(pal_elem->pal, pal,768);

    /*add to list */
    add_to_beginning(*pal_list, pal_elem);
    DBUG_PRINT(3,(LOGF,"add_pal: adding palette,ref=%d \n",
                  pal_elem->ref));

  done:
    return ret_value;
} /* add_pal() */

/*------------------------------------------------------------------------- 
 NAME
       pal_ann2html - Iterator function to print annotation of palette
 DESCRIPTION
       Iterator function to print annotation of palette.
 RETURNS
       Nothing
--------------------------------------------------------------------------*/
void
pal_ann2html(void *dptr, /* palette element */
             void *args)
{
    tan_st *pal_ann = dptr;
    lvar_st *l_vars = args;

    if (pal_ann != NULL)
      {
          if (pal_ann->ann != NULL)
              prhtml_3str(l_vars->hfp,"<PRE>",pal_ann->ann,"</PRE>");
          else
              print_label(l_vars->hfp,"<PRE> -- </PRE>");
      }
} /* pal_ann2html() */


/*------------------------------------------------------------------------
 NAME
    pr_pallabel - Iterator fcn to print labels of palette
 DESCRIPTION
     Iterator fcn to print palette labels to html tbl.
 RETURNS
     Nothing
-------------------------------------------------------------------------*/
void
pr_pallabel(void *dptr, /* raster element */
             void *args)
{
    Generic_list *pal_data_label = dptr;
    lvar_st *l_vars = args;
    intn nlabels = 0;

    nlabels = num_of_objects(*pal_data_label);

    if (nlabels > 0)
       { 
       perform_on_list(*pal_data_label,pal_ann2html,l_vars->hfp);
       }
    else
       fprintf(l_vars->hfp," -- ");
}

/*------------------------------------------------------------------------
 NAME
    pr_paldesc - Iterator fcn to print descriptions of raster
 DESCRIPTION
     Iterator fcn to print raster descriptions to html tbl.
 RETURNS
     Nothing
-------------------------------------------------------------------------*/
void   
pr_paldesc(void *dptr,
             void *args)
{
    Generic_list *pal_data_desc = dptr;
    lvar_st *l_vars = args;
    intn ndescs = 0;
    intn j;
    tan_st *pal_ann =NULL;

    ndescs = num_of_objects(*pal_data_desc);

    /* for descriptions dont print them out, list them */

    if (ndescs > 0)
      {  
      print_label(l_vars->hfp, BGNUL);
      for (j =0; j < ndescs; j++)
        {
         pal_ann = next_in_list(*pal_data_desc);
         if (pal_ann != NULL)
         fprintf(l_vars->hfp,"<LI><A HREF=\"%s\"> desc%d </A></LI>",obj_href(pal_ann->ann_tag,pal_ann->ann_ref,pal_ann->index,l_vars),j);
        }
      print_label(l_vars->hfp, ENDUL);
      }
    else
      fprintf(l_vars->hfp," -- ");
}      
 
/*------------------------------------------------------------------------
 NAME
    pr_palette - Iterator fcn to print palette
 DESCRIPTION
     Iterator fcn to print palette to html tbl.
 RETURNS
     Nothing
-------------------------------------------------------------------------*/
void
pr_palette(void *dptr,
             void *args)
{
    tpal_st *pal_elem = dptr;
    lvar_st *l_vars = args;
 
    fprintf(l_vars->hfp, " <IMG SRC=\"%s%s?%s!hdfref;tag=%d,ref=%d,s=%d\"> ", l_vars->h_env->script_name,l_vars->h_env->path_info,l_vars->f_name, (int32) DFTAG_IP8, pal_elem->ref,0);
}

/*------------------------------------------------------------------------- 
 NAME
       set_lone_pal - Iterator fcn to set this is a lone palette 
 DESCRIPTION
       Iterator fcn to set this is a lone palette given the ref of the 
       palette.
 RETURNS
       Nothing
--------------------------------------------------------------------------*/
void
set_lone_pal(void *dptr, /* palette element from list */
             void *args /* ref to match against */ )
{
    tpal_st *pal_elem = dptr;
    uint16 *ref = args;

    if (pal_elem->ref ==  *ref)
        pal_elem->lone_pal = 0; /* no */
} /* set_lone_pal() */

/*------------------------------------------------------------------------- 
 NAME
     lone_pal - Iterator fcn to indicate whether this is a lone palette  
 DESCRIPTION
     Iterator fcn to indicate whether this is a lone palette  
 RETURNS
     1->SUCCEED, 0->FAIL
--------------------------------------------------------------------------*/
int
lone_pal(void *dptr, /* palette element to inquire */
         void *args)
{
    tpal_st *pal_elem = dptr;

    if (pal_elem->lone_pal)
        return 1;
    else
        return 0;
} /* lone_pal() */

/*------------------------------------------------------------------------- 
 NAME
    free_pal  - frees palette element 
 DESCRIPTION
    Frees palette element       
 RETURNS
    Nothing
--------------------------------------------------------------------------*/
void
free_pal(void *dptr,  /* palette element to free */
         void *args)
{
    tpal_st *pal_elem = dptr;

    if (pal_elem != NULL)
        free(pal_elem);
} /* free_pal() */

/*------------------------------------------------------------------------- 
 NAME
    cleanup_pal - frees palette lists
 DESCRIPTION
    Fress up the palette lists
 RETURNS
    Nothing
--------------------------------------------------------------------------*/
void
cleanup_pal(lvar_st *l_vars)
{
    ENTER(2,"cleanup_pal");
    /* Palette list */
    if (l_vars->pal_list != NULL)
      { /* free elements first */
          DBUG_PRINT(1,(LOGF,"cleanup_pal: destroy pal_list=%d\n", 
                        num_of_objects(*(l_vars->pal_list))));
          perform_on_list(*(l_vars->pal_list),free_pal,NULL);        
          /* destory list itself */
          destroy_list(l_vars->pal_list);

          free(l_vars->pal_list);
          l_vars->pal_list = NULL;
      }
    EXIT(2,"cleanup_pal");
} /* cleanup_pal() */

/*------------------------------------------------------------------------- 
 NAME
       do_pals - process palettes in file and add to global list
 DESCRIPTION
       Process palettes in file and add to global list
 RETURNS
       SUCCEED/FAIL
--------------------------------------------------------------------------*/
int
do_pals(char *fname,   /* file name */
        lvar_st *l_vars)
{
    int32  count;     /* number of palettes */
    int32  i;         /* loop variable */
    uint16 ref;       /* reference number of paletter */
    int32  status;    /* flag */
    char   pal[768];   /* palette */
    FILE   *h_fp = l_vars->hfp;
    int    ret_value = SUCCEED;

    ENTER(2,"do_pals");

    /* Get number of palettes in file */
    count = DFPnpals(fname);

    if(count < 1) 
      {
          ret_value = SUCCEED;
          goto done;
      }

    /* initialize Palette lists */
    if(l_vars->pal_list == NULL)
      {
          /* allocate list to hold atributes */
          if ((l_vars->pal_list = HDmalloc(sizeof(Generic_list))) == NULL)
            {
                gateway_err(l_vars->hfp,"do_pals: failed to allocate space for pals list\n",0);
                ret_value = FAIL; 
                goto done;
            }   

          /* initialize list */
          initialize_list(l_vars->pal_list);
      }

    /* For each palette in the file */
    for(i = 0; i < count; i++) 
      {
          if ((status = DFPgetpal(fname, pal)) == FAIL)
            {
                gateway_err(h_fp,"do_pals: failed getting palette",0);
                ret_value = FAIL;
                goto done;
            }

          /* Save last reference number */
          if ((ref = DFPlastref()) < 1)
            {
                gateway_err(h_fp,"do_pals: failed getting last ref for palette",0);
                ret_value = FAIL;
                goto done;
            }

          /* add palette to list */
          if (add_pal(l_vars->pal_list,ref,pal,l_vars) == FAIL)
            {
                gateway_err(h_fp,"do_pals: failed to add palette to list \n",0);
                ret_value = FAIL;
                goto done;
            }

      } /* end for count */

  done:
    /* failure */
    if (ret_value == FAIL)
      {
          cleanup_pal(l_vars);
      }
    EXIT(2,"do_pals");
    return ret_value;
} /* do_pals */

/*------------------------------------------------------------------------- 
 NAME
       dump_pals -  print out the info for Palettes in the file 
 DESCRIPTION
       print out the info for Palettes in the file 
 RETURNS
       SUCCEED/FAIL
--------------------------------------------------------------------------*/
int
dump_pals(char *fname,    /* file name */
          lvar_st *l_vars)
{
   
    FILE  *h_fp;
    int32 count;        /* number of palettes */
    int32 i;            /* loop variable */
    tpal_st *pal_elem = NULL;
    tan_st  *pal_ann = NULL;
    Generic_list *lone_pals = NULL; /* list of lone palettes */
    Generic_list *pal_data_label = NULL;
    Generic_list *pal_data_desc = NULL;
    int nlabels = 0;
    int ndescs  = 0;
    int ret_value = SUCCEED;
    /* Array of Descriptive Labels for Table Header Columns */
    char *hdr[MAXHDRCOLS] = { RI4, RI5, RI6, RI8};
    /* Number of Descriptive Labels for Table Header Columns */
    int numofhdrs =4;
    

    ENTER(2,"dump_pals");
    /* Get number of Lone palettes in file */
    if(l_vars->pal_list != NULL)
      {
          /* initialize Palette lists */
          if(lone_pals == NULL)
            {
                /* allocate list to hold atributes */
                if ((lone_pals = HDmalloc(sizeof(Generic_list))) == NULL)
                  {
                      gateway_err(l_vars->hfp,"dump_pals: failed to allocate space for pals list\n",0);
                      ret_value = FAIL; 
                      goto done;
                  }   
            }
          *lone_pals = all_such_that(*(l_vars->pal_list),lone_pal,NULL);
          count = num_of_objects(*lone_pals);
      }
    else
      {
          ret_value = SUCCEED;
          goto done;
      }

    /* how many lone paletes do we have */
    if(count < 1) 
      {
          ret_value = SUCCEED;
          goto done;
      }


        h_fp = l_vars->hfp;

    prhtml_title(h_fp,PAL1,count,PAL2,PAL3,PAL4,NULL);

    print_label(h_fp, BGNUL);

    DBUG_PRINT(1,(LOGF, "dump_pals: There are %d lone palettes in this file :\n", count));


          /*  Add table label header strings to list, then print the list */

          print_label(h_fp, HTMLBGNTBL);
          prhtml_tblhdr(h_fp,RI8,hdr,numofhdrs);
 

    /* For each palette in the file */
    for(i = 0; i < count; i++) 
      {
          /* get palette from list */
          pal_elem = next_in_list(*lone_pals);

          /* get palettes labels and descriptions if any */
          if ((pal_data_label = get_data_labels(fname, DFTAG_IP8, pal_elem->ref, h_fp,l_vars)) == NULL)
            {
                gateway_err(h_fp,"dump_pals: error getting data labels",0);
                ret_value = FAIL;
                goto done;
            }

          if ((pal_data_desc = get_data_descs(fname, DFTAG_IP8, pal_elem->ref, h_fp,l_vars)) == NULL)
            {
                gateway_err(h_fp,"dump_pals: error getting data descs",0);
                ret_value = FAIL;
                goto done;
            }
          
            print_label(h_fp, HTMLTBLROW2);
            prhtml_tblentry(h_fp, (void *)(l_vars), (void *)pal_data_label, pr_pallabel);
            prhtml_tblentry(h_fp, (void *)(l_vars),  (void *)pal_data_desc, pr_paldesc);
            prhtml_tblentry(h_fp, (void *)(l_vars), " 256 by 3 ",pr_tblstr);
            prhtml_tblentry(h_fp, (void *)(l_vars), (void *)pal_elem, pr_palette);
            print_label(h_fp,HTMLTBLROWEND);

	/* regular cleanup of labels/descs if any */
          if (pal_data_label != NULL)
            {
                perform_on_list(*(pal_data_label),free_an,NULL);        
                destroy_list(pal_data_label);
                HDfree(pal_data_label);
                pal_data_label = NULL;
            }
          if (pal_data_desc != NULL)
            {
                perform_on_list(*(pal_data_desc),free_an,NULL);        
                destroy_list(pal_data_desc);
                HDfree(pal_data_desc);
                pal_data_desc = NULL;
            }

      } /* for loop for each palette */

    print_label(h_fp, HTMLENDTBL);


  done:
    /* failure */
    if (ret_value == FAIL)
      {
      }

    /* regular cleanup */
    if (lone_pals != NULL)
      {
          /* destory list itself */
          destroy_list(lone_pals);
          free(lone_pals);
      }

    /* regular cleanup */
    if (pal_data_label != NULL)
      {
          perform_on_list(*(pal_data_label),free_an,NULL);        
          destroy_list(pal_data_label);
          HDfree(pal_data_label);
          pal_data_label = NULL;
      }
    if (pal_data_desc != NULL)
      {
          perform_on_list(*(pal_data_desc),free_an,NULL);        
          destroy_list(pal_data_desc);
          HDfree(pal_data_desc);
          pal_data_desc = NULL;
      }

    EXIT(2,"dump_pals");
    return ret_value;
} /* dump_pals */
