/****************************************************************************
 * NCSA HDF                                                                 *
 * Software Development Group                                               *
 * National Center for Supercomputing Applications                          *
 * University of Illinois at Urbana-Champaign                               *
 * 605 E. Springfield, Champaign IL 61820                                   *
 *                                                                          *
 * For conditions of distribution and use, see the accompanying             *
 * COPYING file.                                                            *
 *                                                                          *
 ****************************************************************************/
#ifdef RCSID
static char RcsId[] = "@(#)sdb_sds.c,v 1.40 1996/05/09 20:58:11 georgev Exp";
#endif

/* sdb_sds.c,v 1.40 1996/05/09 20:58:11 georgev Exp */

/*
   LIST Functions
   --------------
     add_sds     - Add SDS to List
     add_attr    - add attriubte to SDS attibute list
     sd_set_lone - Iterator fcn to set this SDS as a Lone SDS
     attr_cmp    - Iterator fcn to compare againse Attrbutes in list given name 
     print_attr  - Debuging, Iterator to print attribute
     sd_ann2html - Iterator fcn to print annotation of SDS
     sds2htmlform - Iterator fcn to print SDS as a html form
     attr2html   - Iterator fcn to print attriubte of  SDS
     pr_sdsdim   - Iterator fcn to print SDS dimension
     pr_sdssize - Iterator fcn to print sds size
     pr_sdssize - Iterator fcn to print sds size
     pr_sdslabel - Iterator fcn to print labels of sds
     pr_sdsdesc - Iterator fcn to print descriptions of sds
     print_sds   - Debuging, Iterator to print SDS info
     var_cmp     - Iterator fcn to find SDS element given name of SDS
     ndg_cmp     - Iterator fcn to find SDS given NDG ref
     is_regvar   - Iterator fcn to indicate whether this is a regular variable
     is_dimvar   - Iterator fcn to indicate whether this is a dimension variable
     is_lone     - Iterator fcn to indicate whether this is a Lone SDS
     free_attr   - frees attribute element 
     free_sds    - Iterator fcn frees SDS element and its attribute list
     cleanup_sds - frees SDS list

   Regular Functions
   -----------------
     readSds  -  Get ready to convert current SDS to the image if possible
     sdsImage - convert current SDS to the image if possible
     do_sds   -  This process SDS and puts them into lists 
     dump_sd  - dump info on SDS given ref of NDG
     dump_sds - Print out the info for data sets in the file. 

 */

#include "sdb.h"
#include "sdb_sds.h"
#include "sdb_attr.h"
#include "sdb_ann.h"
#include "sdb_util.h"
#include "sdb_img.h"
#include "sdsdump.h"
#include "glist.h"
#include "htmlstrings.h"
#include "html_util.h"
#include "string.h"

/*--------------------------------------------------------------------
 NAME
      add_sds - Add SDS to List
 DESCRIPTION
      Adds the SDS to the list along with relevant info.
 RETURNS
      SUCCEED/FAIL
---------------------------------------------------------------------*/
int
add_sds(Generic_list *sds_list,  /* SDS List */
        char *name,              /* name of SDS */
        int var_type,            /* dimension/regular variable */
        int32 rank,              /* rank */
        int32 *dimsizes,         /* dimension array */
        int32 nattrs,            /* number of attributes */
        int32 nt,                /* number type */
        int index,               /* index of this SDS in file */
        int32 ndg_ref,           /* ref of NDG for this SDS */
        uint16 tag,              /* tag, should be DFTAG_NDG except for netCDF*/
        Generic_list *attr_list, /* attribute LIST */
        lvar_st *l_vars)
{
    int i;
    tsds_st *sds_elem = NULL;
    int     ret_value = SUCCEED;

    /* allocate space for SDS element */
    if ((sds_elem = (tsds_st *)HDmalloc(sizeof(tsds_st))) == NULL)
      {
          gateway_err(l_vars->hfp,"add_sds: failed to allocate space for sds_elem\n",0);
          ret_value = FAIL; 
          goto done;
      }   

    /* allocate space for SDS name */
    if ((sds_elem->name = (char *)HDmalloc(sizeof(char)*(HDstrlen(name)+1))) == NULL)
      {
          gateway_err(l_vars->hfp,"add_sds: failed to allocate space for sds_elem->name\n",0);
          ret_value = FAIL; 
          goto done;
      }   
    HDstrcpy(sds_elem->name, name); /* copy name over */

    /* fill in othe relevant info */
    sds_elem->rank    = rank;
    sds_elem->nt      = nt;
    sds_elem->index   = index;
    sds_elem->ndg_ref = ndg_ref;
    sds_elem->tag     = tag;
    sds_elem->lone    = 1; /* yes lone is default */

    /* we assume dimesinos are <= MAXDIM for now
     *  MAXDIM is defined in sdb_sds.h, current value =50  -lpd-*/
    if(rank <= MAXDIM){
        for (i = 0; i < rank; i++)
             sds_elem->dimsizes[i] = dimsizes[i];
    }

    sds_elem->var_type = var_type;
    sds_elem->nattrs   = nattrs;
    sds_elem->tatrlist = attr_list;

    /* add element to SDS list */
    add_to_beginning(*sds_list, sds_elem);
    DBUG_PRINT(3,(LOGF,"add_sds: adding sds,name=%s \n",
                  sds_elem->name));

  done:
    return ret_value;
}/* add_sds() */

/*--------------------------------------------------------------------
 NAME
     add_attr  - add attriubte to SDS attibute list
 DESCRIPTION
     Add attriubte to SDS attibute list
 RETURNS
     SUCCEED/FAIL
---------------------------------------------------------------------*/
int
add_attr(Generic_list *attr_list, /* attribute list of SDS */
         char *name,              /* name of attribute */
         char *value_str,         /* value of attribue */
         lvar_st *l_vars)
{
    tatr_st *attr_elem = NULL;
    int     ret_value = SUCCEED;

    /* allocate space for attribute element */
    if ((attr_elem = (tatr_st *)HDmalloc(sizeof(tatr_st))) == NULL)
      {
          gateway_err(l_vars->hfp,"add_attr: failed to allocate space for attr_elem \n",0);
          ret_value = FAIL; 
          goto done;
      }   

    /* allocate space for name of attribute */
    if ((attr_elem->name = (char *)HDmalloc(sizeof(char)*(HDstrlen(name)+1))) == NULL)
      {
          gateway_err(l_vars->hfp,"add_attr: failed to allocate space for attr_elem->name\n",0);
          ret_value = FAIL; 
          goto done;
      }   
    HDstrcpy(attr_elem->name, name); /* copy name */

    /* allocate space for value of attribute */
    if ((attr_elem->value_str = (char *)HDmalloc(sizeof(char)*(HDstrlen(value_str)+1))) == NULL)
      {
          gateway_err(l_vars->hfp,"add_attr: failed to allocate space attr_elem->value_str\n",0);
          ret_value = FAIL; 
          goto done;
      }   
    HDstrcpy(attr_elem->value_str, value_str); /* copy value */

    /* add to attibute list */
    add_to_beginning(*attr_list, attr_elem);
    DBUG_PRINT(3,(LOGF,"add_attr: adding attribute,name=%s \n",
                  attr_elem->name));

  done:
    /* failure */
    if (ret_value == FAIL)
      {
          if (attr_elem != NULL)
            {
                if (attr_elem->name != NULL)
                    HDfree(attr_elem->name);
                if (attr_elem->value_str != NULL)
                    HDfree(attr_elem->value_str);

                HDfree(attr_elem);
            }
      }

    return ret_value;
} /* add_attr() */

/* ------------------------------------------------------------------------
 NAME
     sd_set_lone - Iterator fcn to set this SDS as a Lone SDS
 DESCRIPTION
      Iterator fcn to set this SDS as a Lone SDS given the NDG ref.
 RETURNS 
      Nothing
--------------------------------------------------------------------------*/
void
sd_set_lone(void *dptr, /* SDS element */
            void *args  /* NDG ref of SDS */)
{
    tsds_st *sds_elem = dptr;
    int32 *ref = args;

    if(sds_elem->ndg_ref == *ref)
        sds_elem->lone = 0;
}/* sd_set_lone */

/*--------------------------------------------------------------------
 NAME
     attr_cmp - Iterator fcn to compare againse Attrbutes in list given name 
 DESCRIPTION
     Iterator fcn to compare againse Attributes in list given name of 
     Attrbute to find.
 RETURNS
     1->SUCCEESS, 0->FAIL
---------------------------------------------------------------------*/
int
attr_cmp(void *dptr, /* Attriute  element */
         void *args  /* name of Attribute to find */)
{
    tatr_st *attr_elem = dptr;
    char *name = args;

    if(!HDstrcmp(attr_elem->name,name))
        return 1;
    else
        return 0;
} /* attr_cmp() */

/*--------------------------------------------------------------------
 NAME
     print_attr - Debuging, Iterator to print attribute
 DESCRIPTION
     Debuging, Iterator to print attribute
 RETURNS
     Nothing;
---------------------------------------------------------------------*/
void
print_attr(void *dptr, 
           void *args)
{
    tatr_st *attr_elem = dptr;

    DBUG_PRINT(3,(LOGF,"print_attr: attribute->name=%s, value=%s \n",
                  attr_elem->name,attr_elem->value_str));
} /* print_attr() */

/*--------------------------------------------------------------------
 NAME
     sd_ann2html - Iterator fcn to print annotation of SDS
 DESCRIPTION
     Iterator fcn to print annotation of SDS
 RETURNS
     Nothing
---------------------------------------------------------------------*/
void
sd_ann2html(void *dptr, /* SDS element */
            void *args)
{
    tan_st *sd_ann = dptr;
    lvar_st *l_vars = args;

    if (sd_ann != NULL)
      {
          if (sd_ann->ann != NULL)
              prhtml_3str(l_vars->hfp,"<PRE>",sd_ann->ann,"</PRE>");
	  else
              print_label(l_vars->hfp,"<PRE> -- </PRE>");

      }
} /* sd_ann2html() */
/*--------------------------------------------------------------------
 NAME
     pr_sdsdim - Iterator fcn to print sds dimension 
 DESCRIPTION
     Iterator fcn to print sds dimension in a
     html tbl entry.
 RETURNS
     Nothing;
---------------------------------------------------------------------*/
void
pr_sdsdim(void *dptr,  /* SDS element */
           void *args)
{
        int j;
        tsds_st *sds_elem = dptr;
        lvar_st *l_vars = args;
 
    /* Print each dimension size in HTML*/
   fprintf(l_vars->hfp, "[");
   for(j = 0; j < sds_elem->rank; j++)
      {
         if (j == 0)
            fprintf(l_vars->hfp, "%d", sds_elem->dimsizes[j]);
         else
            fprintf(l_vars->hfp, ", %d", sds_elem->dimsizes[j]);
       } /* for loop */
            fprintf(l_vars->hfp, "]");
 
      DBUG_PRINT(1,(LOGF, "pr_sdsdim: DFKNTsize(nt)=%d.\n", DFKNTsize(sds_elem->nt | DFNT_NATIVE)));
}


/*--------------------------------------------------------------------
 NAME
     pr_sdssize - Iterator fcn to print sds size
 DESCRIPTION
     Iterator fcn to print sds size in a
     html tbl entry.
 RETURNS 
     Nothing;
---------------------------------------------------------------------*/
void
pr_sdssize(void *dptr,  /* SDS element */
           void *args)
{
        int j,dsize;
        tsds_st *sds_elem = dptr;
        lvar_st *l_vars = args; 
 
   dsize = 1;
   for(j = 0; j < sds_elem->rank; j++)
      dsize = dsize * sds_elem->dimsizes[j];
 
      /* print size  */
   fprintf(l_vars->hfp, "<pre> %d </pre>", (DFKNTsize(sds_elem->nt)*dsize));

      DBUG_PRINT(1,(LOGF, "pr_sdssize:dsize=%d.\n", dsize));
 
}
/*------------------------------------------------------------------------
 NAME
    pr_sdslabel - Iterator fcn to print labels of sds
 DESCRIPTION
     Iterator fcn to print sds labels to html tbl.
 RETURNS 
     Nothing
-------------------------------------------------------------------------*/
void
pr_sdslabel(void *dptr, /* raster element */
             void *args)
{
    Generic_list *sd_data_label = dptr;
    lvar_st *l_vars = args;
    intn nlabels = 0;
 
    nlabels = num_of_objects(*sd_data_label);
    DBUG_PRINT(1,(LOGF, "dump_sd: found %d labels for data,tag=%d \n",nlabels,DFTAG_NDG));
 
    if (nlabels > 0)
       { 
       perform_on_list(*sd_data_label,sd_ann2html,l_vars->hfp);
       } 
    else
       fprintf(l_vars->hfp," -- ");
}

/*------------------------------------------------------------------------
 NAME
    pr_sdsdesc - Iterator fcn to print descriptions of sds
 DESCRIPTION
     Iterator fcn to print sds descriptions to html tbl.
 RETURNS 
     Nothing
-------------------------------------------------------------------------*/
void
pr_sdsdesc(void *dptr,
             void *args)
{
    Generic_list *sd_data_desc = dptr;
    lvar_st *l_vars = args;
    intn ndescs = 0;
    intn j;
    tan_st *sd_ann =NULL;
 
    ndescs = num_of_objects(*sd_data_desc);
    
    DBUG_PRINT(1,(LOGF, "dump_sd: found %d descs for NDG,tag=%d \n",ndescs,DFTAG_NDG 
));
 
   /* for descs show links instead of descs themselves */
         
    if (ndescs > 0)
      {  
      print_label(l_vars->hfp, BGNUL);
      for (j =0; j < ndescs; j++)
        {
         sd_ann = next_in_list(*sd_data_desc);
         if (sd_ann != NULL)
         fprintf(l_vars->hfp,"<LI><A HREF=\"%s\"> desc%d </A></LI>",obj_href(sd_ann->ann_tag,sd_ann->ann_ref,sd_ann->index,l_vars),j);
        }
      print_label(l_vars->hfp, ENDUL);
      }  
    else 
      fprintf(l_vars->hfp," -- ");
}
   
/*------------------------------------------------------------------------
 NAME
     sds2htmlform - Iterator fcn to print sds as a html select form
 DESCRIPTION
     Iterator fcn to print sds as a html select form
 RETURNS
     Nothing
-------------------------------------------------------------------------*/
void 
sds2htmlform(Generic_list list, FILE *fout, int numitems, char *itemlabel)
{
    int i=0;
    tsds_st *dim_elem= NULL;
                            
    for (i = 0; i < numitems; i++)
       { 

        dim_elem = next_in_list(list);
        if(i== 0)
            fprintf(fout, "<OPTION SELECTED VALUE=\";tag=%d,ref=%d,s=%d\"> %s\n", 
                                                DFTAG_NDG,dim_elem->ndg_ref,0,dim_elem->name);
        else
           fprintf(fout, "<OPTION VALUE=\";tag=%d,ref=%d,s=%d\"> %s",DFTAG_NDG,dim_elem->ndg_ref,0 ,dim_elem->name);
         }
    
   DBUG_PRINT(3,(LOGF,"sds2htmlform tag=%d,ref=%d,s=%s:\n",DFTAG_NDG,dim_elem->ndg_ref,dim_elem->name));
 
 
} /* sds2htmlform() */

/*--------------------------------------------------------------------
 NAME
     attr2html - Iterator fcn to print attriubte of  SDS
 DESCRIPTION
     Iterator fcn to print attriubte of  SDS
 RETURNS
     Nothing
---------------------------------------------------------------------*/
void
attr2html(void *dptr, /* attribute element */
          void *args)
{
    tatr_st *attr_elem = dptr;
    FILE *fout = args;

    fprintf(fout, "<TR ALIGN=\"left\" VALIGN=\"top\"><TD> %s </TD> <TD> <pre>%s</pre> </TD></TR> \n",
            attr_elem->name, attr_elem->value_str);

} /* attr2html() */

/*--------------------------------------------------------------------
 NAME
     print_sds - Debuging, Iterator to print SDS info
 DESCRIPTION
     Debuging, Iterator to print SDS info
 RETURNS
     Nothing
---------------------------------------------------------------------*/
void
print_sds(void *dptr, /* SDS elment */
          void *args)
{
    int i;
    tsds_st *sds_elem = dptr;

    DBUG_PRINT(3,(LOGF,"print_sds: sds->name=%s, rank=%d, var_type=%d ",
                  sds_elem->name,sds_elem->rank, sds_elem->var_type));
    DBUG_PRINT(3,(LOGF,", nattrs=%d ",
                  sds_elem->nattrs));
    DBUG_PRINT(3,(LOGF,"dimsizes[ "));
    for(i = 0; i < sds_elem->rank; i++) 
      {
          if (i == 0)
              DBUG_PRINT(3,(LOGF, "%d", sds_elem->dimsizes[i]));
          else
              DBUG_PRINT(3,(LOGF, ", %d", sds_elem->dimsizes[i]));
      } /* for loop */
    DBUG_PRINT(3,(LOGF,"]\n"));

    if(sds_elem->tatrlist != NULL)
      {
          perform_on_list(*sds_elem->tatrlist,print_attr,NULL);
      }
} /* print_sds() */

/*--------------------------------------------------------------------
 NAME
     var_cmp - Iterator fcn to find SDS element given name of SDS
 DESCRIPTION
     Iterator to find SDS element given name of SDS
 RETURNS
     1->SUCCESS, 0->FAIL
---------------------------------------------------------------------*/
int
var_cmp(void *dptr, /* SDS element */
        void *args  /* name of SDS to find */)
{
    tsds_st *sds_elem = dptr;
    char    *name     = args;

    if(!HDstrcmp(sds_elem->name,name))
        return 1;
    else
        return 0;
} /* var_cmp() */

/*--------------------------------------------------------------------
 NAME
      ndg_cmp  - Iterator fcn to find SDS given NDG ref
 DESCRIPTION
      Iterator fcn to find SDS given NDG ref
 RETURNS
     1->SUCCESS, 0->FAIL
---------------------------------------------------------------------*/
int
ndg_cmp(void *dptr, /* SDS element */
        void *args  /* NDG ref to find */)
{
    tsds_st *sds_elem = dptr;
    int32   *ndg_ref  = args;

    if(sds_elem->ndg_ref == *ndg_ref)
        return 1;
    else
        return 0;
}/* ndg_cmp() */

/*--------------------------------------------------------------------
 NAME
      is_regvar - Iterator fcn to indicate whether this is a regular variable
 DESCRIPTION
      Iterator fcn to indicate whether this is a regular variable
 RETURNS
      1-> SUCCESS, 0->FAIL
---------------------------------------------------------------------*/
int
is_regvar(void *dptr, /* SDS element */
          void *args)
{
    tsds_st *sds_elem = dptr;

    if(!sds_elem->var_type)
        return 1;
    else
        return 0;
} /* is_reg_var() */

/*--------------------------------------------------------------------
 NAME
     is_dimvar - Iterator fcn to indicate whether this is a dimension variable
 DESCRIPTION
     Iterator fcn to indicate whether this is a dimension variable
 RETURNS
     1->SUCCESS, 0->FAIL
---------------------------------------------------------------------*/
int
is_dimvar(void *dptr, /* SDS element */
          void *args)
{
    tsds_st *sds_elem = dptr;

    if(sds_elem->var_type)
        return 1;
    else
        return 0;
} /* is_dimvar() */

/*--------------------------------------------------------------------
 NAME
     is_lone - Iterator fcn to indicate whether this is a Lone SDS
 DESCRIPTION
     Iterator fcn to indicate whether this is a Lone SDS
 RETURNS
     1->SUCCESS, 0->FAIL
---------------------------------------------------------------------*/
int
is_lone(void *dptr, /* SDS element */ 
        void *args)
{
    tsds_st *sds_elem = dptr;

    if(sds_elem->lone)
        return 1;
    else
        return 0;
} /* is_lone() */

/*--------------------------------------------------------------------
 NAME
     free_attr - frees attribute element 
 DESCRIPTION
     frees attribute element 
 RETURNS
     Nothing
---------------------------------------------------------------------*/
void
free_attr(void *dptr, /* attribute element to free */
          void *args)
{
    tatr_st *attr_elem = dptr;

    if (attr_elem != NULL)
      {
          if(attr_elem->name != NULL)
              free(attr_elem->name);       /* name */
          if(attr_elem->value_str != NULL)
              free(attr_elem->value_str);  /* value string */
          free(attr_elem); /* structure itself */
      }
} /* free_attr() */

/*--------------------------------------------------------------------
 NAME
      free_sds - Iterator fcn frees SDS element and its attribute list
 DESCRIPTION
      Iterator fcn frees SDS element and its attribute list
 RETURNS
      Nothing
---------------------------------------------------------------------*/
void
free_sds(void *dptr,  /* SDS element to free */
         void *args)
{
    tsds_st *sds_elem = dptr;
    
    if (sds_elem != NULL)
      {
          /* free allocated spce from name */
          if (sds_elem->name != NULL)
              free(sds_elem->name);
            
          /* free attribute list */
          if (sds_elem->tatrlist != NULL)
            {
                perform_on_list(*(sds_elem->tatrlist),free_attr,NULL);
                destroy_list(sds_elem->tatrlist);
                free(sds_elem->tatrlist);
            }

          /* free structure itself */
          free(sds_elem);
      }
} /* free_sds() */

/*--------------------------------------------------------------------
 NAME
     cleanup_sds - frees SDS list
 DESCRIPTION
     frees SDS list
 RETURNS
     Nothing
---------------------------------------------------------------------*/
void
cleanup_sds(lvar_st *l_vars)
{
    ENTER(2,"cleanup_sds");
    /* destroy dimension variable list first */
    if(l_vars->dimvar_list != NULL)
      {
          DBUG_PRINT(1,(LOGF,"cleanup_sds: destroy dimvar_list=%d\n", 
                        num_of_objects(*(l_vars->dimvar_list))));
          destroy_list(l_vars->dimvar_list);
          free(l_vars->dimvar_list);
          l_vars->dimvar_list = NULL;
      }

    /* next the regular variable list */
    if(l_vars->regvar_list != NULL)
      {
          DBUG_PRINT(1,(LOGF,"cleanup_sds: destroy regvar_list=%d\n", 
                        num_of_objects(*(l_vars->regvar_list))));
          destroy_list(l_vars->regvar_list);
          free(l_vars->regvar_list);
          l_vars->regvar_list = NULL;
      }

    /* now the SDS list */
    if (l_vars->sds_list != NULL)
      { /* free elements first */
          DBUG_PRINT(1,(LOGF,"cleanup_sds: destroy sds_list=%d\n", 
                        num_of_objects(*(l_vars->sds_list))));
          perform_on_list(*(l_vars->sds_list),free_sds,NULL);        
          /* destory list itself */
          destroy_list(l_vars->sds_list);

          free(l_vars->sds_list);
          l_vars->sds_list = NULL;
      }

    /* now the global attribute list */
    if (l_vars->gattr_list != NULL)
      { /* free elements first */
          DBUG_PRINT(1,(LOGF,"cleanup_sds: destroy gattr_list=%d\n", 
                        num_of_objects(*(l_vars->gattr_list))));
          perform_on_list(*(l_vars->gattr_list),free_attr,NULL);        
          /* destory list itself */
          destroy_list(l_vars->gattr_list);

          free(l_vars->gattr_list);
          l_vars->gattr_list = NULL;
      }
    EXIT(2,"cleanup_sds");
}/* cleanup_sds() */


/*------------------------------------------------------------------------ 
 NAME
       readSds -  Get ready to convert current SDS to the image if possible
 DESCRIPTION
       Get ready to convert current SDS to the image if possible
 RETURNS
       SUCCEED if conversion is finished otherwise FAIL
-------------------------------------------------------------------------*/
int
readSds(char *fileName,  /* file name */
        int32 ref,       /* NDG ref of SDS */
        lvar_st *l_vars)
{
    int32 i;       /* loop variable */
    int   width;       /* width dimension of image */
    int   height;       /* height dimesnion of image */
    int   rank;
    int   plane;
    int   ip = 0;  /* pallete flag  */
    tsds_st *sds_elem=NULL;
    FILE  *h_fp;
    int   ret_value = SUCCEED;
 
    ENTER(2,"readSds");
    /* get file identifier */
    h_fp = l_vars->hfp;
 
    /* get ready to read SDS from HDF file */
    DBUG_PRINT(1,(LOGF,"hdf_ref = %s \n",l_vars->hdf_ref ));
    DBUG_PRINT(1,(LOGF,"ref = %d \n",ref ));

    /* l_vars->sds_list hold the generic information of the current SDS */
    /* find dataset matchine hdf_ref */
    sds_elem = first_that(*(l_vars->sds_list),ndg_cmp,&ref);

    if (sds_elem == NULL) 
      {
          DBUG_PRINT(1,(LOGF, "Not found DataSet->ref %d\n", ref));
          ret_value = FAIL;
          goto done;
      }

    rank = sds_elem->rank;
    if ((rank < 2) || (rank>3))
      {
          ret_value = FAIL;
          goto done;
      }

    width = sds_elem->dimsizes[rank-1];
    height= sds_elem->dimsizes[rank-2];

    if ((width*height)<= 0)
      {
          ret_value = FAIL;
          goto done;
      }
  
    plane =1;
    for (i=0; i<(rank-2); i++)
        plane *= sds_elem->dimsizes[i];

    DBUG_PRINT(1,(LOGF,"rank = %d \n",rank ));
    DBUG_PRINT(1,(LOGF,"width = %d \n",width ));
    DBUG_PRINT(1,(LOGF,"height = %d \n",height ));
    DBUG_PRINT(1,(LOGF,"plane = %d \n",plane ));
 
    fprintf(h_fp, "<HR>\n");

    /* Print Image header stuff in HTML  */
    fprintf(h_fp, "<H2>Images from the current SDS(%s)</H2>\n",sds_elem->name); 

    fprintf(h_fp, "<UL>\n");

    /* Print out image info. in HTML  */

    if ((rank>2) && (plane>1)) /* using first plane of the image */
        fprintf(h_fp,"<LI> This image has <b>%d</b> planes. This <A HREF=\"%s%s?%s!sdbref;ref=%d,s=%d,plane=%d\"><IMG SRC=\"%s%s?%s!sdbref;ref=%d,s=%d,plane=%d\"> </A> is the plane %d of the image with dimensions %d by %d\n",
                plane, 
                l_vars->h_env->script_name,l_vars->h_env->path_info,l_vars->f_name,
                ref, 0,l_vars->plane,l_vars->h_env->script_name, 
                l_vars->h_env->path_info,l_vars->f_name, ref, 1,l_vars->plane, \
                l_vars->plane, width, height);

    else
        fprintf(h_fp,"<LI> This image  <A HREF=\"%s%s?%s!sdbref;ref=%d,s=%d,plane=1\"><IMG SRC=\"%s%s?%s!sdbref;ref=%d,s=%d,plane=1\"></A> has dimensions %d by %d\n", 
                l_vars->h_env->script_name,l_vars->h_env->path_info,l_vars->f_name,
                ref, 0, l_vars->h_env->script_name, 
                l_vars->h_env->path_info,l_vars->f_name, ref, 1, width, height);

    if(width > hdfImageSize || height > hdfImageSize)
        fprintf(h_fp, "  (the image has been subsampled for display)");
    fprintf(h_fp, ".  ");
            
    if(ip) 
        fprintf(h_fp, "There is also a palette associated with this image.\n");

    
    if ((rank>2)&&(plane>1))  { /* dimensions > 2)  */
		
        if (l_vars->plane<plane)
            ++l_vars->plane;
    
        /*
          fprintf(l_vars->hfp,"<LI>To see the next plane of the image, <b>click <A HREF=\"%s%s?%s!sdbplane;plane=%d\"> here </A> </b>\n", l_vars->h_env->script_name,l_vars->h_env->path_info,l_vars->f_name,l_vars->plane);
  
          */

        fprintf(l_vars->hfp,"<LI>To see the next plane of the image, <b>click <A HREF=\"%s%s?%s!sdbplane;ref=%d,plane=%d\"> here </A> </b>\n", l_vars->h_env->script_name,l_vars->h_env->path_info,l_vars->f_name,ref,l_vars->plane);


        fprintf(l_vars->hfp,"<LI>To preview all the planes winthin the image, <b>click <A HREF=\"%s%s?%s!sdbview;ref=%d,start=%d,end=%d\"> here </b> </A> \n", l_vars->h_env->script_name,l_vars->h_env->path_info,l_vars->f_name,ref,1,plane);


        fprintf(l_vars->hfp,"<LI>To preview the group of the planes winthin the image, please enter: \n");
		
        fprintf(l_vars->hfp,"<FORM METHOD=\"POST\" \n");
        fprintf(l_vars->hfp,"ACTION=\"%s%s\">\n",
                l_vars->h_env->script_name,
                l_vars->h_env->path_info);

        fprintf(l_vars->hfp, "<INPUT TYPE=\"hidden\" NAME=\"SDB_IMGFLAG\" VALUE=\"%d\">\n",SDSVIEW);

        fprintf(l_vars->hfp, "<INPUT TYPE=\"hidden\" NAME=\"f_name\" VALUE=\"%s\">\n",l_vars->f_name);
        fprintf(l_vars->hfp, "<INPUT TYPE=\"hidden\" NAME=\"SDB_REF\" VALUE=\"%d\">\n",ref);
        fprintf(l_vars->hfp, "<INPUT TYPE=\"hidden\" NAME=\"SDB_PLANE\"  VALUE=\"%d\">\n",plane);

        fprintf(l_vars->hfp, "<UL>");
	
        fprintf(l_vars->hfp, "<LI>Starting plane:");
        fprintf(l_vars->hfp, "<INPUT NAME=\"SDB_START\" VALUE=%d>\n", 1);
		
        fprintf(l_vars->hfp, "<LI>Ending   plane:");
        fprintf(l_vars->hfp, "<INPUT NAME=\"SDB_END\" VALUE=%d>\n", plane);
	          
        fprintf(l_vars->hfp, "</UL>");
	
        prhtml_submitbutton(l_vars->hfp,"",SDS5, OUT3, 0,NULL, 0, NULL, 0);
        print_label(l_vars->hfp,"</FORM>");


    }
    
    /* end of image */
    fprintf(l_vars->hfp, "</UL>\n");

  done:

    EXIT(2,"readSds()");
    return ret_value;
} /* readSds() */


/*------------------------------------------------------------------------ 
 NAME
       sdsImage - convert current SDS to the image if possible
 DESCRIPTION
       convert current SDS to the image if possible
 RETURNS
       SUCCEED if conversion is finished otherwise FAIL
-------------------------------------------------------------------------*/
int
sdsImage(int32 sdsid, 
         char *imageData , 
         int plane)
{
    /* some of sds variable */
    int32 naxis,naxes[15];
    int32 nelmt;
    int32 datatype;
    int32 numPlane; /* total number of the plane */
    int32 start[5];
    int32 stride[5];
    int32 edge[5];
    float *buffer=NULL,*buf=NULL;
    int   datasize, i;
    char  sdsName[25];
    int32 nattrs;
    /* pointer of various types */
    union {
      char      *cp;    /* Byte or char */
      short     *sp;    /* short */
      uint16    *usp;   /* unsigned short */
      int       *ip;    /* integer */
      int32     *uip;   /* unsigned int */
      float32   *fp;    /* float */
      float64   *dp;    /* double */
    } databuf;
    char *ptr;

    int  status = 0;

    ENTER(2,"sdsImage");
    
    /* read the required primary array keywords  */
    status = SDgetinfo(sdsid, sdsName, &naxis, naxes, &datatype, &nattrs);

    if (status == FAIL)
        return FAIL;

    if ((naxis <= 1) || (naxis >=4))
        return (FAIL);

    numPlane = 1;
    for (i=0; i<(naxis-2); i++)
        numPlane *= naxes[i];

    if (plane > numPlane)
        return(FAIL);
    
    /*          w          *       h    */
    nelmt =  naxes[naxis-1]*naxes[naxis-2];

    buffer =  (float *)HDmalloc(nelmt*sizeof(float));

    if (buffer == NULL)
        /* out of memory */
        return(FAIL);

    buf = buffer;

    DBUG_PRINT(1,(LOGF, "Data Number Type: %d\n", datatype));

    switch(datatype) {
   
    case DFNT_UCHAR:
    case DFNT_CHAR:   /* char */
    case DFNT_INT8:   /* signed integer */
    case DFNT_UINT8:  /* unsigned char  */
        datasize = nelmt * sizeof(char);
        break;
    case DFNT_UINT16: /* unsigned short */
    case DFNT_INT16:  /* short */
        datasize = nelmt * sizeof(int16);
        break;
    case DFNT_UINT32:
    case DFNT_INT32:  /* integer */
        datasize = nelmt * sizeof(int32);
        break;
    case DFNT_FLOAT32:  /* float */
        datasize = nelmt * sizeof(float);
        break;
    case DFNT_FLOAT64:  /* double */
        datasize = nelmt * sizeof(double);
        break;
    default: 
        DBUG_PRINT(1,(LOGF, "unsupported Data Number Type: %d\n", datatype));
        return(FAIL);
    }

    databuf.cp = ptr = (char *)HDmalloc(datasize);

    if (databuf.cp == NULL)
        /* out of memory */
        return(FAIL);

    if (naxis == 3) {  /* 3-d image */
        start[0] = (plane-1);
        start[1] = 0;
        start[2] = 0;

        stride[0] = 1;
        stride[1] = 1;
        stride[2] = 1;

        edge[0] = 1;                /* z(length) */
        edge[1] = naxes[naxis-2];   /* y(length) */
        edge[2] = naxes[naxis-1];   /* x(length) */
    }
    else {      
        start[0] = 0;
        start[1] = 0;
  
        stride[0] = 1;
        stride[1] = 1;
 
        edge[0] = naxes[naxis-2];   /* y(length) */
        edge[1] = naxes[naxis-1];   /* x(length) */
    }


    DBUG_PRINT(1,(LOGF, "plane= %d\n", plane));
 
    status = SDreaddata(sdsid, start, NULL, edge, databuf.cp);

    DBUG_PRINT(1,(LOGF, "status = %d\n", status));

    if (status == FAIL)
        return FAIL;

    switch (datatype){
 
    case DFNT_CHAR:
    case DFNT_INT8: {

        for (i=0; i<nelmt; i++)
            *buf++ = *databuf.cp++;
        /*                    x               y                      */   
        convert2image(buffer, naxes[naxis-1], naxes[naxis-2], (uint8 *)imageData);
        break;
    } 

    case DFNT_UCHAR:             /* byte/char */
    case DFNT_UINT8:
        /*get image data by plane */

        for (i=0; i<nelmt; i++)
            imageData[i] = databuf.cp[i];

        break;

    case DFNT_INT16: {  /* short */

        for (i=0; i<nelmt; i++)
            *buf++ = *databuf.sp++;
        /*                    x               y                      */   
        convert2image(buffer, naxes[naxis-1], naxes[naxis-2], (uint8 *)imageData);
        break;
    } 
    case DFNT_UINT16: {  /* unsigned short */

        for (i=0; i<nelmt; i++)
            *buf++ = *databuf.usp++;
        /*                    x               y                      */   
        convert2image(buffer, naxes[naxis-1], naxes[naxis-2], (uint8 *)imageData);
        break;
    }
    case DFNT_INT32: {
		
        for (i=0; i<nelmt; i++)
            *buf++ = *databuf.ip++;

        /*                    x               y                      */   
        convert2image(buffer, naxes[naxis-1], naxes[naxis-2], (uint8 *)imageData);

        break;
    }
    case DFNT_UINT32: {
		
        for (i=0; i<nelmt; i++)
            *buf++ = *databuf.uip++;

        /*                    x               y                      */   
        convert2image(buffer, naxes[naxis-1], naxes[naxis-2], (uint8 *)imageData);

        break;
    }
 
    case DFNT_FLOAT32: {
	
        for (i=0; i<nelmt; i++) 
            *buf++ = *databuf.fp++;
	
        /*                    x               y                      */   
        convert2image(buffer, naxes[naxis-1], naxes[naxis-2], (uint8 *)imageData);

        break;
    }
 
    case DFNT_FLOAT64: { /* double */
     
        for (i=0; i<nelmt; i++)
            *buf++ = *databuf.dp++;

        /*                    x               y                      */   
        convert2image(buffer, naxes[naxis-1], naxes[naxis-2], (uint8 *)imageData);
	
        break;
    }
    default:
        DBUG_PRINT(1,(LOGF, "Unknown data number type: %d\n", datatype));
        return FAIL;
    }
     
    HDfree(ptr); 
    HDfree(buffer);

    EXIT(2,"sdsImage");  
    return(SUCCEED);
} /* sdsImage() */

/*--------------------------------------------------------------------
 NAME
     do_sds -  This process SDS and puts them into lists 
 DESCRIPTION
     this process SDS and puts them into lists 
 RETURNS
     SUCCEED/FAIL
---------------------------------------------------------------------*/
int
do_sds(char *fname, 
       lvar_st *l_vars)
{
    int32 i, j;         /* loop variables */
    int32 fid = FAIL;   /* file handle */
    int32 sds = FAIL;   /* Data set handle */
    int32 dsets;        /* number of data sets */
    int32 ngattr;       /* number of global attributes */
    int32 ndattr;       /* number of attriubtes of data set */
    int32 status;       /* flag */
    char  name[512];    /* sds name */
    int32 nt;           /* Number type */
    int32 dimsizes[MAXDIM]; /* array for dimension sizes */
    int32 rank;         /* rank of data set */
    int32 count;        /* number of attributes */
    int32 ndg_ref;      /* ref of NDG */
    uint16 sd_tag = DFTAG_NDG;      /* DFTAG_NDG except for netCDF */
    FILE *h_fp   = l_vars->hfp;
    Generic_list  *tsds_attrlist = NULL; /* pointer to attribute list */
    int           ret_value = SUCCEED;

    ENTER(2,"do_sds");
    DBUG_PRINT(1,(LOGF," l_vars->display_mode=%d\n",l_vars->display_mode)); 
    DBUG_PRINT(3,(LOGF," before l_vars->hdf_ref=%s\n",l_vars->hdf_ref)); 

    /* initialize SDS and global attribute lists */
    if(l_vars->sds_list == NULL)
      {
          /* allocate list to hold atributes */
          if ((l_vars->sds_list = HDmalloc(sizeof(Generic_list))) == NULL)
            {
                gateway_err(l_vars->hfp,"do_sds: failed to allocate space for SDS list\n",0);
                ret_value = FAIL; 
                goto done;
            }   

          /* initialize list */
          initialize_list(l_vars->sds_list);
      }
    if(l_vars->gattr_list == NULL)
      {
          /* allocate list to hold atributes */
          if ((l_vars->gattr_list = HDmalloc(sizeof(Generic_list))) == NULL)
            {
                gateway_err(l_vars->hfp,"do_sds: failed to allocate space for SDS list\n",0);
                ret_value = FAIL; 
                goto done;
            }   

          /* initialize list */
          initialize_list(l_vars->gattr_list);
      }

    DBUG_PRINT(1,(LOGF," fname=%s\n",fname)); 

    /* Open file for reading */
    if ((fid = SDstart(fname, DFACC_RDONLY)) == FAIL)
      {
          gateway_err(l_vars->hfp,"do_sds: opening file ",0);
          ret_value =  FAIL;
          goto done;
      }

    /* Find number of datasets and global attributes */
    if ((status = SDfileinfo(fid, &dsets, &ngattr)) == FAIL)
      {
          gateway_err(l_vars->hfp,"do_sds: getting file info",0);
          ret_value = FAIL;
          goto done;
      }

    DBUG_PRINT(1,(LOGF,"do_sds: dsets=%d\n",dsets)); 
    DBUG_PRINT(1,(LOGF,"do_sds:  ngattr=%d\n",ngattr)); 

    /* Check on the number of datasets and global attriubtes we found */
    if(dsets + ngattr < 1) 
      {
          DBUG_PRINT(1,(LOGF,"do_sds:  nothing to process for sds\n")); 
          ret_value = SUCCEED;
          goto done;
      }

    /* if we have datasets process them */
    if (dsets)
      {
          /* process each dataset */
          for (i=0; i < dsets; i++) 
            {
                int dimvar;

                /* select variable */
                if ((sds=SDselect(fid, i))==FAIL)
                  {
                      gateway_err(h_fp,"do_sds: failed to select dataset \n",0);
                      ret_value = FAIL;
                      goto done;
                  }

                /* Is is this a dimension variable */
                dimvar=(SDiscoordvar(sds))?1:0;

                if (l_vars->do_netcdf)
                  {
                      ndg_ref = i;    /* we use the index for netCDF */
                      sd_tag  = 6666; /* fake tag for netCDF */
                  }
                else
                  {
                      /* ref of NDG */
                      if ((ndg_ref=SDidtoref(sds))==FAIL)
                        {
                            gateway_err(h_fp,"do_sds: failed to get ndg ref of dataset \n",0);
                            ret_value = FAIL;
                            goto done;
                        }
                  }
                
                /* Get info about variable like name, rank, dimension sizes,
                 * number type and number of attributes */
                if((status = SDgetinfo(sds, name, &rank, dimsizes, &nt, &ndattr)) == FAIL)
                  {
                      SDendaccess(sds);
                      gateway_err(h_fp,"do_sds: getting dataset info \n",0);
                      ret_value = FAIL;
                      goto done;
                  }

                DBUG_PRINT(1,(LOGF, "do_sds: variable(ndg=%d) %d %s rank %d \n", 
                              ndg_ref,i, name, rank));

                /* deal with variable attributes */
                if (ndattr)
                  {
                      DBUG_PRINT(1,(LOGF, "        variable has %d attributes\n",ndattr));
                      /* allocate list to hold atributes */
                      if ((tsds_attrlist = HDmalloc(sizeof(Generic_list))) == NULL)
                        {
                            SDendaccess(sds);
                            gateway_err(l_vars->hfp,"do_sds: failed to allocate space for attribute list\n",0);
                            ret_value = FAIL; 
                            goto done;
                        }   

                      /* initialize attriube list of this SDS */
                      initialize_list(tsds_attrlist);

                      /* for each attribute of dataset */
                      for(j = 0; j < ndattr; j++) 
                        {
                            char *val_str = NULL; /* value of attribute */
                            char t_name[512];
                            int32 t_nt;
                            int32 acount;

                            /* Get attribute info */
                            if ((status = SDattrinfo(sds, j, t_name, &t_nt, &acount)) == FAIL)
                              {
                                  SDendaccess(sds);
                                  gateway_err(h_fp,"do_sds: failed to get attribute info for SDS \n",0);
                                  ret_value = FAIL;
                                  goto done;
                              }

                            /* Convert attribue object into a string */
                            if ((val_str = get_attribute(sds, j, t_nt, acount,l_vars)) == NULL)
                              {
                                  gateway_err(h_fp,"do_sds: failed to convert attribute into a string\n",0);
                                  continue;
                              }

                            DBUG_PRINT(1,(LOGF, "do_sds:   %s : %s \n", t_name, val_str));

                            /* add attribute to attribute list */
                            if (add_attr(tsds_attrlist,t_name,val_str,l_vars) == FAIL)
                              {
                                  SDendaccess(sds);
                                  gateway_err(h_fp,"do_sds: failed to add attribute to list \n",0);
                                  ret_value = FAIL;
                                  goto done;
                              }

                            /* We free space allocated in "get_attribute" */
                            if(val_str != NULL)
                                HDfreespace((void *)val_str);
                        } /* for naddtr */
                  }
                else /* no attributes */
                    tsds_attrlist = NULL;

                /* add sds w/ attribute list to SDS list */
                if (add_sds(l_vars->sds_list,name,dimvar,rank,dimsizes,ndattr,nt,i,
                            ndg_ref,sd_tag,tsds_attrlist,l_vars) == FAIL)
                  {
                      SDendaccess(sds);
                      gateway_err(h_fp,"do_sds: failed to add sds to list \n",0);
                      ret_value = FAIL;
                      goto done;
                  }

                DBUG_PRINT(1,(LOGF, "do_sds: added variable %s to list \n", name));
                /* end acces to variable */
                SDendaccess(sds);

            } /* end for dsets */
          DBUG_PRINT(1,(LOGF, "do_sds: done adding variables to list \n"));
      } /* end if dsets */

    /* process global attributes if any */
    if(ngattr)
      {
          DBUG_PRINT(1,(LOGF, "do_sds: adding global attributes to list \n"));
          /* for each global attribute */
          for(i = 0; i < ngattr; i++) 
            {
                char *valstr; /*attribute string */

                /* Get attribute info */
                if ((status = SDattrinfo(fid, i, name, &nt, &count)) == FAIL)
                  {
                      gateway_err(h_fp,"do_sds: failed to get global attributes",0);
                      ret_value = FAIL;
                      goto done;
                  }

                /* Covert attribute object into a string */
                if ((valstr = get_attribute(fid, i, nt, count,l_vars)) == NULL)
                  {
                      gateway_err(h_fp,"do_sds: failed to convert global attribute into a string\n",0);
                      continue;
                  }
                DBUG_PRINT(1,(LOGF, "      %s : %s \n", name, valstr));

                /* add attribute to attribute list */
                if (add_attr(l_vars->gattr_list,name,valstr,l_vars) == FAIL)
                  {
                      gateway_err(h_fp,"do_sds: failed to add attribute to global list \n",0);
                      ret_value = FAIL;
                      goto done;
                  }

                /* We free space allocated in "get_attribute" */
                if(valstr != NULL)
                    HDfreespace((void *)valstr);
            } /* end for loop for each global attribute */
      } /* end if ngattr */


    /* get list of dimension variables */
    if(l_vars->dimvar_list == NULL)
      {
          /* allocate list to hold atributes */
          if ((l_vars->dimvar_list = HDmalloc(sizeof(Generic_list))) == NULL)
            {
                gateway_err(l_vars->hfp,"do_sds: failed to allocate space for SDS list\n",0);
                ret_value = FAIL; 
                goto done;
            }   
      }
    *(l_vars->dimvar_list) = all_such_that(*(l_vars->sds_list),is_dimvar,NULL);

    /* get list of regular variables */
    if(l_vars->regvar_list == NULL)
      {
          /* allocate list to hold atributes */
          if ((l_vars->regvar_list = HDmalloc(sizeof(Generic_list))) == NULL)
            {
                gateway_err(l_vars->hfp,"do_sds: failed to allocate space for SDS list\n",0);
                ret_value = FAIL; 
                goto done;
            }   
      }
    *(l_vars->regvar_list) = all_such_that(*(l_vars->sds_list),is_regvar,NULL);

  done:
    /* failure */
    if (ret_value == FAIL)
      {
          cleanup_sds(l_vars);
      }

    /* cleanup */
    if (fid != FAIL)
        SDend(fid);

    EXIT(2,"do_sds");
    return ret_value;
} /* do_sds() */

/*--------------------------------------------------------------------
 NAME
      dump_sd - dump info on SDS given ref of NDG
 DESCRIPTION
      dump info on SDS given ref of NDG
 RETURNS
      SUCCEED/FAIL
---------------------------------------------------------------------*/
int
dump_sd(char *fname,     /* name of file */
        int32 sdref,     /* ref of ndg of SDS */
        lvar_st *l_vars)
{
    int j;
    int32 dsets;    /* number of data sets */
    int32 ngattr;   /* number of global attributes */
    int32 fid = -1;
    int32 sds = -1;
    int32 display_mode =1;
    int ndimvars = 0;
    int nregvars = 0;
    tsds_st      *sds_elem = NULL;
    Generic_list *sd_data_label = NULL;
    Generic_list *sd_data_desc = NULL;
    tan_st       *sd_ann = NULL;
    int    nlabels, ndescs;
    uint16 ref;
    FILE   *h_fp = l_vars->hfp;
    int    ret_value = SUCCEED;

    /* Array of Descriptive Labels for Table Header Columns */
    char *hdr[MAXHDRCOLS] = { SDS6,SDS7,SDS8,SDS9,SDS10};

    /* Number of Descriptive Labels in hdr array (above) */
    int numofhdrs =5;

    /* Array of Radio Button Labels for Selecting Output Mode */
    /* first name in list is the name of "selected", default radio button */
    char *outputbuttons[MAXHDRCOLS] = { OUT15, OUT15, OUT11, OUT12, OUT13, OUT14 };
 
    /* Number of Output Buttons */
    int numradiobuttons =6;

 

    /* Declaration for the Dib palettes inserted by lpd */
    FILE *pal = NULL;
    int npals, k, n;
    char tmp_html[MAX_STRING_LEN];
    char l[MAX_STRING_LEN];
    char w[MAX_STRING_LEN];


    ENTER(2,"dump_sd");
    DBUG_PRINT(1,(LOGF,"dump_sd: l_vars->display_mode=%d\n",l_vars->display_mode)); 
    DBUG_PRINT(3,(LOGF,"dump_sd: before l_vars->hdf_ref=%s\n",l_vars->hdf_ref)); 
   

    /* We key of the display_mode */
    display_mode = l_vars->display_mode;
    l_vars->hdf_ref++;
    DBUG_PRINT(3,(LOGF,"dump_sds: before l_vars->hdf_ref=%s\n",l_vars->hdf_ref)); 

    /* number of SDS and global attributes */
    if (l_vars->sds_list != NULL)
        dsets = num_of_objects(*(l_vars->sds_list));
    else
        dsets = 0;
    DBUG_PRINT(1,(LOGF, "dump_sd: there are %d data sets \n", dsets));
    if (l_vars->gattr_list != NULL)
        ngattr = num_of_objects(*(l_vars->gattr_list));
    else
        ngattr = 0;
    DBUG_PRINT(1,(LOGF, "dump_sd: there are %d global attriubes \n", ngattr));

    /* count home many dimension and regualr variables there are */
    if (l_vars->dimvar_list != NULL)
        ndimvars = num_of_objects(*(l_vars->dimvar_list));
    else
        ndimvars = 0;
    DBUG_PRINT(1,(LOGF, "dump_sd: there are %d dimension variables \n", ndimvars));
    if (l_vars->regvar_list != NULL)
        nregvars = num_of_objects(*(l_vars->regvar_list));
    else
        nregvars = 0;
    DBUG_PRINT(1,(LOGF, "dump_sd: there are %d regular variables \n", nregvars));

    DBUG_PRINT(1,(LOGF," fname=%s\n",fname)); 

    /* find correct sds */
    sds_elem = first_that(*(l_vars->sds_list),ndg_cmp,&sdref);

    if (sds_elem != NULL)
      {
          DBUG_PRINT(1,(LOGF, "dump_sd: found DataSet->ndg_ref %d\n", sdref));
                /* are we subsetting */
                if (!l_vars->sub_s)
                  { /* not subsetting dataset */
                      /* prepare HTML table for SDS info */

			print_label(h_fp, HTMLBGNTBL);
                        prhtml_tblhdr(h_fp,"",hdr,numofhdrs);
                        print_label(h_fp, HTMLTBLROW2);
                        prhtml_tblentry (h_fp, (void *)(l_vars), sds_elem->name, pr_tblstr);
                        prhtml_tblentry (h_fp, (void *)(l_vars), &(sds_elem->rank), pr_tblnum);
                        prhtml_tblentry (h_fp, (void *)(l_vars), (void *)(sds_elem), pr_sdsdim);
                        prhtml_tblentry (h_fp, (void *)(l_vars), (get_type(sds_elem->nt)),pr_tblstr);
                        prhtml_tblentry (h_fp, (void *)(l_vars), (void *)(sds_elem), pr_sdssize );
                        print_label(h_fp, HTMLTBLROWEND);
                        print_label(h_fp, HTMLENDTBL);

                      /* set ndg ref */
                      ref = (uint16)sdref;

                      /* netCDF files dont' have annotations */
                      if (!l_vars->do_netcdf)
                        {
                            /* prepare Table for labesl/descs */

    			    /* Array of Descriptive Labels for Table Header Columns */
    		            hdr[0]= SDS10;
    		            hdr[1]= SDS11;

    		 	   /* Number of Descriptive Labels in hdr array (above) */
                           print_label(h_fp, HTMLBGNTBL);
    			   numofhdrs =2;
                           prhtml_tblhdr(h_fp,"",hdr,numofhdrs);

                            /* get data labels and descriptions if any */
                            if ((sd_data_label = get_data_labels(fname, DFTAG_NDG, ref, h_fp,l_vars)) == NULL)
                              {
                                  gateway_err(h_fp,"dump_sd: error getting data labels",0);
                                  ret_value = FAIL;
                                  goto done;
                              }
                            nlabels = num_of_objects(*sd_data_label);

                            DBUG_PRINT(1,(LOGF, "dump_sd: found %d labels for data,tag=%d,ref=%d \n",
                                          nlabels,DFTAG_NDG,ref));    


                            if ((sd_data_desc = get_data_descs(fname, DFTAG_NDG, ref, h_fp,l_vars)) == NULL)
                              {
                                  gateway_err(h_fp,"dump_sd: error getting data descs",0);
                                  ret_value = FAIL;
                                  goto done;
                              }
                            ndescs = num_of_objects(*sd_data_desc);

                            DBUG_PRINT(1,(LOGF, "dump_sd: found %d descs for NDG,tag=%d,ref=%d \n",
                                          ndescs,DFTAG_NDG,ref));    

                            print_label(h_fp, HTMLTBLROW2);
                            prhtml_tblentry(h_fp,(void *)(l_vars),(void *)sd_data_label, pr_sdslabel);
                            prhtml_tblentry(h_fp,(void *)(l_vars),(void *)sd_data_desc, pr_sdsdesc); 
                            print_label(h_fp, HTMLTBLROWEND);
                            print_label(h_fp, HTMLENDTBL);

                        } /* if not netCDF */

                      /* deal with attributes */
                      if (sds_elem->nattrs)
                        {
                            reset_to_beginning(*(sds_elem->tatrlist));

                            /* Array of Descriptive Labels for Table Header Columns */
                            hdr[0]= ATT2;
                            hdr[1]= ATT3;

                           /* Number of Descriptive Labels in hdr array (above)*/
                           
                           print_label(h_fp, HTMLBGNTBL);
                           numofhdrs =2;
                           prhtml_tblhdr(h_fp,"",hdr,numofhdrs);
                           prhtml_wholetbl(h_fp, *(sds_elem->tatrlist),attr2html); 
                           print_label(h_fp, HTMLENDTBL);
                        } /* end if nattrs */


                      /* ====================SDS to image===================*/
                      /* take a sds to be as am image here?  
                         fprintf(h_fp, "<HR>");
                         fprintf(h_fp, "Insert the source code for generating the image from the SDS ");
                         */      
                      readSds(fname, sdref, l_vars);
                      fprintf(h_fp, "<HR>");	      
                      /* ================================================== */

                      /* now the rest which is the FORM for subsetting 
                       *  the SDS */
                      if (sds_elem->rank > 0)
                        {
                      fprintf(h_fp, "<FORM METHOD=\"POST\" ");
                      fprintf(h_fp, "ACTION=\"%s%s\">\n",l_vars->h_env->script_name,
                              l_vars->h_env->path_info);
                      fprintf(h_fp, "<INPUT TYPE=\"hidden\" NAME=\"SD_RANK\" ");
                      fprintf(h_fp, "VALUE=\"%d\">\n", sds_elem->rank);
                      fprintf(h_fp, "<PRE WIDTH=\"%d\">", DOC_WIDTH);
            
                      for (j=0; j < sds_elem->rank; j++) 
                        {
                            fprintf(h_fp, "Dimension %d: <P>\n", j);
                            fprintf(h_fp, "<UL>");
                            fprintf(h_fp, "<LI>starting location :");
                            fprintf(h_fp, "<INPUT NAME=\"SD_DIM_START\" VALUE=%d> <P>\n", 
                                    0); 
                            fprintf(h_fp, "<LI>number of units to skip:");
                            fprintf(h_fp, "<INPUT NAME=\"SD_DIM_STRIDE\" VALUE=%d> <P>\n", 
                                    1);
                            fprintf(h_fp, "<LI>ending location :");
                            fprintf(h_fp, "<INPUT NAME=\"SD_DIM_END\" VALUE=%d> <P>\n", 
                                    sds_elem->dimsizes[j]);
                            fprintf(h_fp, "</UL>");
                        }
                      fprintf(h_fp, "</PRE>");

                      fprintf(h_fp, "\n<b>The following selections are for retrieving as image only</b><p>");
                      fprintf(h_fp, "\nValue scaling methods:<p>");
                      if(sds_elem->nt == DFNT_CHAR || sds_elem->nt == DFNT_CHAR8 || sds_elem->nt == DFNT_INT8 || sds_elem->nt == DFNT_UINT8){
                             fprintf(h_fp, "<INPUT type=\"radio\" NAME=\"scaling\", VALUE=1 checked> None  ");
                             fprintf(h_fp, "<INPUT type=\"radio\" NAME=\"scaling\", VALUE=2> Linear  ");
                      }
                      else
                             fprintf(h_fp, "<INPUT type=\"radio\" NAME=\"scaling\", VALUE=2 checked> Linear  ");

                      fprintf(h_fp, "<INPUT type=\"radio\" NAME=\"scaling\", VALUE=3> Histogram equalization<p>\n");

                      fprintf(h_fp, "Palette:\n<select NAME=palette>");
                   
                      /* check the internal HDF palette */
                      npals = DFPnpals(fname);
                      DBUG_PRINT(1,(LOGF,"number_of_pals =%d, file=%s\n", npals, fname));
                      if(npals == 0)
                            fprintf(h_fp, " <option selected>Black-White");
                      else {
                           for(k=1; k<= npals; k++){
                                   if(k == 1)
                                          fprintf(h_fp, " <option selected>HDF_Internal_%d", k);
                                       else
                                          fprintf(h_fp, " <option>HDF_Internal_%d", k);
                           }
                           fprintf(h_fp, " <option>Black-White");
                      }

                  /* Parse config file, looking for optional palettes */

                      pal=fopen(SDB_CONFIG_FILE, "r");
                      if(pal){
                      while(!(cfg_getline(l,MAX_STRING_LEN,pal))) {
                       ++n;
                       if((l[0] != '#') && (l[0] != '\0')) {
                           cfg_getword(tmp_html,l);

                           if(!strcmp(tmp_html, "Palette")) {
                           cfg_getword(tmp_html,l);
                              if (tmp_html[0] != '\0')
                              fprintf(h_fp, " <option>%s", tmp_html);
                           }
                           }
                           }


                            fclose(pal);
                      }
                      fprintf(h_fp, "</select>");     
                      fprintf(h_fp, "</PRE>");
                      /* end insert -- lpd */

                      prhtml_select_outputtype(h_fp,OUT5,outputbuttons,numradiobuttons);  
                      prhtml_submitbutton(h_fp,OUT1, OUT2, OUT3, DFTAG_NDG, ref,1,l_vars->f_name,3);
                      print_label(h_fp,"</FORM>"); 
                      
		      fprintf(h_fp, "<em> Please make a note of the size of browse");
                      fprintf(h_fp, "data that is being requested.\n");
                      fprintf(h_fp, "An ascii dump usually can run 3-5 times ");
                      fprintf(h_fp, "larger than the binary size.\n </em>");
                      fprintf(h_fp, "<hr>\n");
                        }
                      else if (sds_elem->rank == 0 && l_vars->do_netcdf)
                        {
                            fprintf(h_fp, "This is probably a netCDF scalar variable\n");
                            fprintf(h_fp, "which we are not yet able to dump \n");
                        }
                  }
                else /* we are subsetting */
                  {
                      fprintf(h_fp, "<B>Dataset %s</B>:-<P>\n", sds_elem->name);
                      if (sds_elem->nattrs)
                        {
                            reset_to_beginning(*(sds_elem->tatrlist));
                            /* Array of Descriptive Labels for Table Header Columns */                   
                            hdr[0]= ATT2; 
                            hdr[1]= ATT3;
 
                           /* Number of Descriptive Labels in hdr array (above) */ 
                           print_label(h_fp, HTMLBGNTBL);
                           numofhdrs =2;
                           prhtml_tblhdr(h_fp,"",hdr,numofhdrs);           
                           prhtml_wholetbl(h_fp, *(sds_elem->tatrlist),attr2html);           
                           print_label(h_fp, HTMLENDTBL);
                        
			} /* end if nattrs */

                      else
                          fprintf(h_fp, "It has no attributes.<P>\n");

                      /* Code for subsetting an SDS */
                      if (sds_elem->rank > 0)
                        {
                            int32 *start = NULL;
                            int32 *stride = NULL;
                            int32 *edge = NULL;
                            int32 *left = NULL;
                            int ret, r, end;

                            /* Open file for reading */
                            if ((fid = SDstart(fname, DFACC_RDONLY)) == FAIL)
                              {
                                  gateway_err(l_vars->hfp,"dump_sds: opening file ",0);
                                  ret_value =  FAIL;
                                  goto done;
                              }

                            /* Select dataset "i" */
                            if ((sds = SDselect(fid, sds_elem->index)) == FAIL)
                              {
                                  SDend(fid);
                                  gateway_err(h_fp,"dump_sds: selecting dataset \n",0);
                                  ret_value =  FAIL;
                                  goto done;
                              }

                            DBUG_PRINT(1,(LOGF, "dump_sds: selected dataset \n"));
                            fprintf(h_fp, "Data :<P>\n");
	    
                            start  = (int32 *)HDgetspace(sds_elem->rank * sizeof(int32));
                            stride = (int32 *)HDgetspace(sds_elem->rank * sizeof(int32));
                            edge   = (int32 *)HDgetspace(sds_elem->rank * sizeof(int32));
                            left   = (int32 *)HDgetspace(sds_elem->rank * sizeof(int32));
                            if (start == NULL || stride == NULL || edge == NULL || left == NULL)
                              {
                                  SDendaccess(sds);
                                  SDend(fid);
                                  gateway_err(h_fp,"dump_sds: failed to allocate space \n",0);
                                  ret_value = FAIL;
                                  goto done;
                              }

                            /* Get subsampling values along each dimension */
                            for (r = 0; r < sds_elem->rank; r++)   
                              {
                                  DBUG_PRINT(1,(LOGF, "dump_sds: sd_dim_start[%d]=%d \n",
                                                r,l_vars->sd_dim_start[r]));
                                  DBUG_PRINT(1,(LOGF, "dump_sds: sd_dim_stride[%d]=%d \n",
                                                r,l_vars->sd_dim_stride[r]));
                                  DBUG_PRINT(1,(LOGF, "dump_sds: sd_dim_end[%d]=%d \n",
                                                r,l_vars->sd_dim_end[r]));
                                  start[r] = l_vars->sd_dim_start[r];
                                  if ((stride[r] = l_vars->sd_dim_stride[r])==0)
                                      stride[r] = 1;
                                  if ((end = l_vars->sd_dim_end[r])==0)
                                      left[r] = sds_elem->dimsizes[r] - start[r];
                                  else
                                      left[r] = end - start[r];
                                  edge[r] = 1;
                              }
                            edge[sds_elem->rank-1] = left[sds_elem->rank-1];

                            /* dump subset of SDS */
                            ret = sdsdumpfull(sds, sds_elem->rank, sds_elem->dimsizes, sds_elem->nt, 16, h_fp, 
                                              start, stride, edge, left, l_vars); 
                            if (ret == FAIL)
                              {
                                  SDendaccess(sds);
                                  SDend(fid);
                                  gateway_err(h_fp,"dump_sds: failed to dump sds \n",0);
                                  ret_value = FAIL;
                                  goto done;
                              }

                            fprintf(h_fp, "<P>\n");
                             
                            /* free up space */
                            HDfreespace(start);
                            HDfreespace(stride);
                            HDfreespace(edge);
                            HDfreespace(left);
                            DBUG_PRINT(1,(LOGF, "dump_sds: done dumping dataset \n"));

                            SDendaccess(sds);
                            SDend(fid);
                        } /* end sds_elem->rank */
                  }
      }

  done:
    if (ret_value == FAIL)
      {
      }

    /* cleanup labels and descs for SDS */
    if (sd_data_label != NULL)
      {
          perform_on_list(*(sd_data_label),free_an,NULL);        
          destroy_list(sd_data_label);
          HDfree(sd_data_label);
          sd_data_label = NULL;
      }
    if (sd_data_desc != NULL)
      {
          perform_on_list(*(sd_data_desc),free_an,NULL);        
          destroy_list(sd_data_desc);
          HDfree(sd_data_desc);
          sd_data_desc = NULL;
      }

    EXIT(2,"dump_sd");
    return ret_value;
} /* dump_sd() */

/*--------------------------------------------------------------------
 NAME
       dump_sds - Print out the info for data sets in the file. 
 DESCRIPTION
       Print out the info for data sets in the file. 
       Still nees work on a rewrite to work better with dump_sd()
 RETURNS
       SUCCEED/FAIL
---------------------------------------------------------------------*/
int
dump_sds(char *fname,                /* file name */
         int current_entry,      
         entry entries[MAX_ENTRIES], 
         int num_entries,
         lvar_st *l_vars)
{
    int32 fid;      /* file handle */
    int32 sds;      /* Data set handle */
    int32 dsets;    /* number of data sets */
    int32 ldsets;   /* lone SDS */
    int32 ngattr;   /* number of global attributes */
    int32 j;     /* loop variables */
    int32 display_mode=1; /* There are three display modes:
                             (1) display the scroll boxes containing the 
                             names of scientific datasets in a file,
                             (2) display a form with three fields, namely,
                             "start", "stride", and "end",	
                             (3) display data in the dataset. */
    char *temp;
    FILE *h_fp = l_vars->hfp;
    int ndimvars = 0;
    int nldimvars = 0;
    int nregvars = 0;
    int nlregvars = 0;
    Generic_list  lone_sds;      /* lone SDS list */
    Generic_list  lone_dimvars;  /* lone dimension variable list */
    Generic_list  lone_regvars;  /* lone regular variable list */
    int  ret_value = SUCCEED;

    /* Array of Descriptive Labels for Table Header Columns */
    char *hdr[MAXHDRCOLS] = { SDS6,SDS7,SDS8,SDS9,SDS10};

    /* Number of Descriptive Labels in hdr array (above) */
    int numofhdrs =5;

    ENTER(2,"dump_sds");
    DBUG_PRINT(1,(LOGF,"dump_sds: l_vars->display_mode=%d\n",l_vars->display_mode)); 
    DBUG_PRINT(3,(LOGF,"dump_sds: before l_vars->hdf_ref=%s\n",l_vars->hdf_ref)); 

/* We key off the display_mode */
    display_mode = l_vars->display_mode;
    l_vars->hdf_ref++;
    if(display_mode == 3)
      { /* adjust so that we can easily extract hdf ref */
          if ((temp = (char *)strrchr(l_vars->hdf_ref, ';')) != NULL)   
            {
                temp[0] = '\0';
                temp++;
            }
      }

    /* number of SDS and global attributes */
    if (l_vars->sds_list != NULL)
        dsets = num_of_objects(*(l_vars->sds_list));
    else
        dsets = 0;
    DBUG_PRINT(1,(LOGF, "dump_sds: there are %d data sets \n", dsets));

    /* find lone SDS, lone dimvars, lone regvars */
    if (dsets > 0)
      {
          lone_sds = all_such_that(*(l_vars->sds_list),is_lone,NULL);
          ldsets = num_of_objects(lone_sds);
          DBUG_PRINT(1,(LOGF, "dump_sds: there are %d Lone Data sets \n", ldsets));
      }
    else
        ldsets = 0;

    /* global attributes */
    if (l_vars->gattr_list != NULL)
        ngattr = num_of_objects(*(l_vars->gattr_list));
    else
        ngattr = 0;
    DBUG_PRINT(1,(LOGF, "dump_sds: there are %d global attriubes \n", ngattr));

    /* count home many dimension and regualr variables there are */
    if (l_vars->dimvar_list != NULL)
        ndimvars = num_of_objects(*(l_vars->dimvar_list));
    else
        ndimvars = 0;
    DBUG_PRINT(1,(LOGF, "dump_sds: there are %d dimension variables \n", ndimvars));

    if (ndimvars > 0)
      {
          lone_dimvars = all_such_that(*(l_vars->dimvar_list),is_lone,NULL);
          nldimvars = num_of_objects(lone_dimvars);
          DBUG_PRINT(1,(LOGF, "dump_sds: there are %d Lone dimension variables \n", nldimvars));
      }
    else
        nldimvars = 0;

    if (l_vars->regvar_list != NULL)
        nregvars = num_of_objects(*(l_vars->regvar_list));
    else
        nregvars = 0;
    DBUG_PRINT(1,(LOGF, "dump_sds: there are %d regular variables \n", nregvars));

    if (nregvars > 0)
      {
          lone_regvars = all_such_that(*(l_vars->regvar_list),is_lone,NULL);
          nlregvars = num_of_objects(lone_regvars);
          DBUG_PRINT(1,(LOGF, "dump_sds: there are Lone %d regular variables \n", nlregvars));
      }
    else
        nldimvars = 0;

    DBUG_PRINT(1,(LOGF,"dump_sds: fname=%s\n",fname)); 
    DBUG_PRINT(1,(LOGF,"dump_sds: display_mode=%d\n",display_mode)); 
    DBUG_PRINT(3,(LOGF,"dump_sds: after l_vars->hdf_ref=%s\n",l_vars->hdf_ref)); 

    /*
     * Now to process the objects and generate HTML
     */

    /* Create HTML tile header for data sets */
          if (display_mode==1 && (dsets || ngattr))
            {
                prhtml_title(h_fp,SDS16,ldsets,SDS17,SDS18,SDS19,NULL); 
                prhtml_title(h_fp,NULL,ngattr,SDS20,SDS18,SDS21,NULL); 

                DBUG_PRINT(1,(LOGF, "dump_sds: There are %d Lone datasets and %d global attribute in this file. \n", ldsets, ngattr)); 
            
	    }

    /* Lets deal with datasets first */
    if(dsets) 
      {
          if (display_mode==1 )
            {
                /* This is where we deal with Lone SDS */
                if (ldsets)
                  {
                      /* lets deal with dimension variables first */
                      if (nldimvars) 
                        {
         		    prhtml_form(h_fp, SDS13, l_vars->h_env->script_name,l_vars->h_env->path_info, SDS14, (dsets > 1? 5:2),SDS2,SDS1, l_vars->f_name,2, lone_dimvars, sds2htmlform,nldimvars, NULL);

                        } /* end if ndimvars */

                      /* deal with regular variables */
                      if (nlregvars) 
                        {

         		    prhtml_form(h_fp, SDS15, l_vars->h_env->script_name,l_vars->h_env->path_info, SDS14, (dsets > 1? 5:2),SDS3,SDS1, l_vars->f_name,2, lone_regvars, sds2htmlform,nlregvars, NULL);

                        } /* end if nregvars */
                  } /* if ldsets */

            } /* end if display_mode == 1 */
          else if (display_mode==2 ) 
            {
                tsds_st *sds_elem = NULL;
                Generic_list *sd_data_label = NULL;
                Generic_list *sd_data_desc = NULL;
                int nlabels, ndescs;
                uint16 ref;
                tan_st *sd_ann = NULL;

                h_fp = l_vars->hfp;

                /* find dataset matchine hdf_ref */
                sds_elem = first_that(*(l_vars->sds_list),var_cmp,l_vars->hdf_ref);

                if (sds_elem != NULL)
                  {
                      DBUG_PRINT(1,(LOGF, "dump_sds: found DataSet->hdf_ref %s\n", l_vars->hdf_ref));
                      /* Print name and rank of dataset in HTML */

                      print_label(h_fp, HTMLBGNTBL);
                      prhtml_tblhdr(h_fp,"",hdr,numofhdrs);

                        print_label(h_fp, HTMLTBLROW2);
                        prhtml_tblentry (h_fp, (void *)(l_vars), sds_elem->name, pr_tblstr);
                        prhtml_tblentry (h_fp, (void *)(l_vars), &(sds_elem->rank), pr_tblnum);
                        prhtml_tblentry (h_fp, (void *)(l_vars), (void *)(sds_elem), pr_sdsdim);
                        prhtml_tblentry (h_fp, (void *)(l_vars), (get_type(sds_elem->nt)),pr_tblstr);
                        prhtml_tblentry (h_fp, (void *)(l_vars), (void *)(sds_elem), pr_sdssize );
                        print_label(h_fp, HTMLTBLROWEND);
                        print_label(h_fp, HTMLENDTBL);

                      /* set ndg ref */
                      ref = (uint16)sds_elem->ndg_ref;

                      /* make sure we are not netCDF */
                      if (!l_vars->do_netcdf)
                        { 
                        /* Array of Descriptive Labels for Table Header Columns */                               
                           hdr[0]= SDS11;
                           hdr[1]= SDS12;

                           /* Number of Descriptive Labels in hdr array
(above) */
                           print_label(h_fp, HTMLBGNTBL);
                           numofhdrs =2;
                           prhtml_tblhdr(h_fp,"",hdr,numofhdrs);
                      
		      /* get data labels and descriptions if any */
                      if ((sd_data_label = get_data_labels(fname, DFTAG_NDG, ref, h_fp,l_vars)) == NULL)
                        {
                            gateway_err(h_fp,"dump_sds: error getting data labels",0);
                            ret_value = FAIL;
                            goto done;
                        }
                      nlabels = num_of_objects(*sd_data_label);

                      DBUG_PRINT(1,(LOGF, "dump_sds: found %d labels for data,tag=%d,ref=%d \n",
                                    nlabels,DFTAG_NDG,ref));    

                      if ((sd_data_desc = get_data_descs(fname, DFTAG_NDG, ref, h_fp,l_vars)) == NULL)
                        {
                            gateway_err(h_fp,"dump_sds: error getting data descs",0);
                            ret_value = FAIL;
                            goto done;
                        }
                      ndescs = num_of_objects(*sd_data_desc);

                      DBUG_PRINT(1,(LOGF, "dump_sds: found %d descs for NDG,tag=%d,ref=%d \n",
                                    ndescs,DFTAG_NDG,ref));    

                            print_label(h_fp, HTMLTBLROW2);
                            prhtml_tblentry(h_fp,(void *)(l_vars),(void *)sd_data_label, pr_sdslabel);
                            prhtml_tblentry(h_fp,(void *)(l_vars),(void *)sd_data_desc, pr_sdsdesc); 
                            print_label(h_fp, HTMLTBLROWEND);
                            print_label(h_fp, HTMLENDTBL);

                        } /* end if not netCDF */

                      /* deal with attributes */
                      if (sds_elem->nattrs)
                        {
                            reset_to_beginning(*(sds_elem->tatrlist));
 
                         /* Array of Descriptive Labels for Table Header Columns */                               
                             hdr[0]= ATT2;
                             hdr[1]= ATT3;
                             /* Number of Descriptive Labels in hdr array
(above) */
                             print_label(h_fp, HTMLBGNTBL);
                             numofhdrs =2;
                             prhtml_tblhdr(h_fp,ATT1,hdr,numofhdrs);
                             prhtml_wholetbl(h_fp,*(sds_elem->tatrlist),attr2html);
                             print_label(h_fp, HTMLENDTBL);
                        } /* end if nattrs */


                      /* ====================== SDS to Image ====================*/
                      /* take a sds to be as am image here? 
                      fprintf(h_fp, "<HR>");
                      fprintf(h_fp, "Insert the source code for generating the image from the SDS ");
		      */
                      readSds(fname, sds_elem->ndg_ref, l_vars);
                      fprintf(h_fp, "<HR>");	      
                      /* ================================================== */
             
                      /* now the rest */
                      fprintf(h_fp, "<FORM METHOD=\"POST\" ");
                      fprintf(h_fp, "ACTION=\"%s%s\">\n",l_vars->h_env->script_name,
                              l_vars->h_env->path_info);
                      fprintf(h_fp, "<PRE WIDTH=\"%d\">", DOC_WIDTH);
            
                      for (j=0; j < sds_elem->rank; j++) 
                        {
                            fprintf(h_fp, "Dimension %d: <P>\n", j);
                            fprintf(h_fp, "<UL>");
                            fprintf(h_fp, "<LI>starting location :");
                            fprintf(h_fp, "<INPUT NAME=\"@%s;dim%d,start\" VALUE=%d> <P>\n", 
                                    l_vars->hdf_ref, j, 0); /*"hdf_ref" now contains the name of
                                                              the requested SD. */
                            fprintf(h_fp, "<LI>number of units to skip:");
                            fprintf(h_fp, "<INPUT NAME=\"@%s;dim%d,stride\" VALUE=%d> <P>\n", 
                                    l_vars->hdf_ref, j, 1);
                            fprintf(h_fp, "<LI>ending location :");
                            fprintf(h_fp, "<INPUT NAME=\"@%s;dim%d,end\" VALUE=%d> <P>\n", 
                                    l_vars->hdf_ref, j, sds_elem->dimsizes[j]);
                            fprintf(h_fp, "</UL>");
                        }
                      fprintf(h_fp, "</PRE>");
                      fprintf(h_fp, "To see the data, press this button: \n");
                      fprintf(h_fp, "<INPUT TYPE=\"submit\" VALUE=\"Browse Data\">. \n");
                      fprintf(l_vars->hfp, "<INPUT TYPE=\"hidden\" NAME=\"hdfref\" VALUE=\";tag=%d,ref=%d,s=%d\">\n",
                              DFTAG_NDG,sds_elem->ndg_ref,1);
                      fprintf(h_fp, "<INPUT TYPE=\"hidden\" NAME=\"f_name\" ");
                      fprintf(h_fp, "VALUE=\"%s\">\n", l_vars->f_name);
                      fprintf(h_fp, "<INPUT TYPE=\"hidden\" NAME=\"display_mode\" VALUE=\"%d\">\n",
                              3);
                      fprintf(h_fp, "<INPUT TYPE=\"hidden\" NAME=\"display_type\" VALUE=\"%d\">\n",
                              1);
                      fprintf(h_fp, "</FORM>");
                      fprintf(h_fp, "<em> Please make a note of the size of browse");
                      fprintf(h_fp, "data that is being requested.\n");
                      fprintf(h_fp, "An ascii dump usually can run 3-5 times ");
                      fprintf(h_fp, "larger than the binary size.\n </em>");
                      fprintf(h_fp, "<hr>\n");

                      /* cleanup */
                      if (sd_data_label != NULL)
                        {
                            perform_on_list(*(sd_data_label),free_an,NULL);        
                            destroy_list(sd_data_label);
                            HDfree(sd_data_label);
                            sd_data_label = NULL;
                        }
                      if (sd_data_desc != NULL)
                        {
                            perform_on_list(*(sd_data_desc),free_an,NULL);        
                            destroy_list(sd_data_desc);
                            HDfree(sd_data_desc);
                            sd_data_desc = NULL;
                        }

                  }
                else
                  {
                      DBUG_PRINT(1,(LOGF, "dump_sds: failed to find DataSet->hdf_ref %s\n", l_vars->hdf_ref));
                      goto done;
                  }
            } /* end display_mode == 2 */
          else if (display_mode==3 ) 
            {
                tsds_st *sds_elem = NULL;

                h_fp = l_vars->hfp;

                /* find dataset matchine hdf_ref */
                sds_elem = first_that(*(l_vars->sds_list),var_cmp,l_vars->hdf_ref);

                if (sds_elem != NULL)
                  {
                      DBUG_PRINT(1,(LOGF, "dump_sds: found DataSet->hdf_ref %s\n", l_vars->hdf_ref));
                            fprintf(h_fp, "<B>Dataset %s</B>:-<P>\n", sds_elem->name);
                            if (sds_elem->nattrs)
                              {
                                  reset_to_beginning(*(sds_elem->tatrlist));

                                  hdr[0]= ATT2;
                                  hdr[1]= ATT3;
                                  
			          print_label(h_fp, HTMLBGNTBL);
                                  /* Number of Labels in hdr array (above) */
                                  numofhdrs =2;
                                  prhtml_tblhdr(h_fp,ATT1,hdr,numofhdrs);
                                  prhtml_wholetbl(h_fp,*(sds_elem->tatrlist),attr2html); 
                                  print_label(h_fp, HTMLENDTBL);

                              }
                            else
                                fprintf(h_fp, "It has no attributes.<P>\n");

                            /* Code for subsetting an SDS */
                            if (sds_elem->rank > 0)
                              {
                                  int32 *start, *stride, *edge, *left;
                                  int ret, r, end;

                                  /* Open file for reading */
                                  if ((fid = SDstart(fname, DFACC_RDONLY)) == FAIL)
                                    {
                                        gateway_err(l_vars->hfp,"dump_sds: opening file ",0);
                                        ret_value =  FAIL;
                                        goto done;
                                    }

                                  /* Select dataset "i" */
                                  if ((sds = SDselect(fid, sds_elem->index)) == FAIL)
                                    {
                                        SDend(fid);
                                        gateway_err(h_fp,"dump_sds: selecting dataset \n",0);
                                        ret_value =  FAIL;
                                        goto done;
                                    }

                                  DBUG_PRINT(1,(LOGF, "dump_sds: selected dataset \n"));
                                  fprintf(h_fp, "Data :<P>\n");
	    
                                  start  = (int32 *)HDgetspace(sds_elem->rank * sizeof(int32));
                                  stride = (int32 *)HDgetspace(sds_elem->rank * sizeof(int32));
                                  edge   = (int32 *)HDgetspace(sds_elem->rank * sizeof(int32));
                                  left   = (int32 *)HDgetspace(sds_elem->rank * sizeof(int32));
                                  if (start == NULL || stride == NULL || edge == NULL || left == NULL)
                                    {
                                        SDendaccess(sds);
                                        SDend(fid);
                                        gateway_err(h_fp,"dump_sds: failed to allocate space \n",0);
                                        ret_value = FAIL;
                                        goto done;
                                    }

                                  /* Get subsampling values along each dimension */
                                  for (r = 0; r < sds_elem->rank; r++)   
                                    {
                                        start[r] = atoi(entries[r*3].val);
                                        if ((stride[r] = atoi(entries[r*3+1].val))==0)
                                            stride[r] = 1;
                                        if ((end = atoi(entries[r*3+2].val))==0)
                                            left[r] = sds_elem->dimsizes[r] - start[r];
                                        else
                                            left[r] = end - start[r];
                                        edge[r] = 1;
                                    }
                                  edge[sds_elem->rank-1] = left[sds_elem->rank-1];

                                  /* dump subset of SDS */
                                  ret = sdsdumpfull(sds, sds_elem->rank, sds_elem->dimsizes, sds_elem->nt, 16, h_fp, 
                                                    start, stride, edge, left, l_vars); 
                                  if (ret == FAIL)
                                    {
                                        SDendaccess(sds);
                                        SDend(fid);
                                        gateway_err(h_fp,"dump_sds: failed to dump sds \n",0);
                                        ret_value = FAIL;
                                        goto done;
                                    }

                                  fprintf(h_fp, "<P>\n");
                             
                                  /* free up space */
                                  HDfreespace(start);
                                  HDfreespace(stride);
                                  HDfreespace(edge);
                                  HDfreespace(left);
                                  DBUG_PRINT(1,(LOGF, "dump_sds: done dumping dataset \n"));

                                  SDendaccess(sds);
                                  SDend(fid);
                              } /* end sds_elem->rank */
                  }
                else
                  {
                      DBUG_PRINT(1,(LOGF, "dump_sds:failed to find DataSet->hdf_ref %s\n", l_vars->hdf_ref));
                      goto done;
                  }
            } /* end display_mode == 3 */

      } /* if(dsets) */

    /* 
     * Now Deal with global attributes 
     */
    if (display_mode==1 )  
      {
          if (ngattr) 
            {
                tatr_st *spattr_elem = NULL;
                h_fp = l_vars->hfp;

                /* check for special global attribute "file description" */
                spattr_elem = first_that(*(l_vars->gattr_list),
                                         attr_cmp,"file description");

                if (spattr_elem != NULL)
                  {
                      fprintf(h_fp, "<UL>\n");
                      fprintf(h_fp, "<LI> Attribute <i>%s</i> has the value : <pre>%s</pre> \n", 
                              spattr_elem->name, spattr_elem->value_str);
                      fprintf(h_fp, "</UL>\n");
                  }

                /* print global attriubes in a Table */
                reset_to_beginning(*(l_vars->gattr_list));

		/* Array of Descriptive Labels for Table Header Columns */
                hdr[0]= ATT2;
                hdr[1]= ATT3;
			         
                print_label(h_fp, HTMLBGNTBL);
                /* Number of Descriptive Labels in (above)hdr array */
                numofhdrs =2;

                prhtml_tblhdr(h_fp,ATT1,hdr,numofhdrs);
                prhtml_wholetbl(h_fp,*(l_vars->gattr_list),attr2html); 
                print_label(h_fp, HTMLENDTBL);

            } /* end if (ngattr) */
      } /* end if display_mode == 1*/


  done:
    if (ret_value == FAIL)
      {
      }

    /* destory tmp lone sds list */
    if (dsets)
        destroy_list(&lone_sds);
    if (ndimvars)
        destroy_list(&lone_dimvars);
    if (nregvars)
        destroy_list(&lone_regvars);

    /* normal cleanup */
   
    EXIT(2,"dump_sds");
    return ret_value;
} /* dump_sds */

